/* -------------------------------------------------------------
 * Copyright 2003-2010 Netlogic Microsystems Inc. (“Netlogic”).
 * -------------------------------------------------------------
 * This is a derived work from software originally provided by  
 * the external entity identified below. The licensing terms and  
 * warranties specified in the header of the original work apply 
 * to this derived work.
 * --------------------------#NETL_1#---------------------------
 */
/*
 * Copyright (c) 2005-2008 Analog Devices Inc.
 *
 * Licensed under the GPL-2 or later.
 */
#include <common.h>
#include <asm/arch/cpu.h>
#include <spi.h>

struct spi_slave spi_dev;		/* SPI NOR  */
struct spi_slave spi_dev_nand;	/* SPI NAND */

/* Common Slave Device initialization 
 * routine used by NAND and NOR devices.
 * Returns the Global Data struct for 
 * chosen device-type
 */
struct spi_slave *spi_setup_slave(u32 bus, u32 cs,
		u32 max_hz, u32 mode, u32 device)
{
	volatile u32 *spi_mmio = cpu_io_mmio(0, SPI);
	struct spi_slave* slave = NULL;

	/* SPI Enable Pin Muxing 
	 */
	spi_mmio[SPI_SYSCNTRL] = SPI_SYSCNTRL_ENABLE;

	spi_mmio = spi_mmio + (CS_OFFSET * cs);

	if ((max_hz < CONFIG_SPI_REFCLK/4) &&
		(max_hz > CONFIG_SPI_REFCLK/(64*1024))) {
		spi_mmio[CS0_FDIV] = (CONFIG_SPI_REFCLK/max_hz);
	} else {
		printf("Unsupported SPI frequency %d\n",max_hz);
		goto out;
	}

	/* Threshold reg 
	 */
	spi_mmio[CS0_FIFO_THRESH] = SPI_THRESH_RX(0x1);

	/* SPI Cfg register 
	 */
	spi_mmio[CS0_CONFIG] = (
		SPI_CFG_CPHA(((mode & SPI_CPHA) == 1)) 		|
		SPI_CFG_CPOL(((mode & SPI_CPOL) == 1))    	|
		SPI_CFG_CSPOL(((mode & SPI_CS_HIGH) == 0)) 	|
		SPI_CFG_TXMISO(((mode & SPI_3WIRE) == 1))   |  
		SPI_CFG_TXMOSI(((mode & SPI_3WIRE) == 0))   | 
		SPI_CFG_RXMISO(((1)))   					|
		SPI_CFG_LSBFE(((mode & SPI_LSB_FIRST) == 1))|
		SPI_CFG_CSTODATA((0)));		    

	/* Allocate appropriate global
	 * structures for nor/nand
	 */
	if (device == SPINAND) {
		slave = &spi_dev_nand;
	}
	else {
		slave = &spi_dev;
	}
		

	slave->bus 	= bus;
	slave->cs  	= cs;
out:
	return slave;
}

void spi_free_slave(struct spi_slave *slave)
{
	return;
}

int spi_claim_bus(struct spi_slave *slave)
{
	return 0;
}

void spi_release_bus(struct spi_slave *slave)
{
	return;
}

static void spi_txfifo_enq(volatile u32* spi_mmio,
			   u32* bitlen, u8** data) {

	u32 txfifo_cnt;
	u32 tx_data;

	txfifo_cnt = SPI_TXFIFO_CNT(spi_mmio[CS0_FIFO_WCNT]);

	while ((*bitlen) && (txfifo_cnt < SPI_FIFO_SIZE))
	{
		if (*bitlen <= 8) {
			tx_data = (*data)[0];
			*bitlen = 0;
		} else if (*bitlen <= 16) {
			tx_data = (((*data)[0] << 8) |
					((*data)[1] << 0));
			*bitlen = 0;
		} else if (*bitlen <= 24) {
			tx_data = (((*data)[0] << 16) |
					((*data)[1] << 8) | 
					((*data)[2] << 0));
			*bitlen = 0;
		} else {
			tx_data = (((*data)[0] << 24) |
					((*data)[1] << 16) |
					((*data)[2] << 8)  | 
					((*data)[3] << 0));
			*bitlen = *bitlen - 32;
		}

		spi_mmio[CS0_TXDATA_FIFO] = tx_data;
		(*data) += 4;
		txfifo_cnt++;
	}	
}

static int spi_block_xfer(struct spi_slave* slave,
			u32 bitlen, const void *dout,
			void *din, unsigned long flags) {

	u8* tx = (u8 *) dout;
	u8* rx = (u8 *) din; 
	u32 rx_data, tx_bitlen, rx_bitlen, rxfifo_cnt;

	volatile u32* spi_mmio = cpu_io_mmio(0, SPI) + (slave->cs * CS_OFFSET);

	if (!bitlen)
		return -1;

	tx_bitlen = (tx == NULL) ? 0 : bitlen;
	rx_bitlen = (rx == NULL) ? 0 : bitlen;

	if (tx_bitlen) {
		spi_txfifo_enq(spi_mmio, &tx_bitlen, &tx);
	}

	spi_mmio[CS0_CMD] = (SPI_CMD_TX((tx != NULL)) |
						 SPI_CMD_RX((rx != NULL)) |
						 SPI_CMD_CONT(((flags & SPI_XFER_END) == 0)) |
						 SPI_CMD_XFERCNT((bitlen - 1)));

	while ((tx_bitlen) || (rx_bitlen)) {

		if (rx_bitlen) {

			rxfifo_cnt = SPI_RXFIFO_CNT(spi_mmio[CS0_FIFO_WCNT]);

			while (rxfifo_cnt) {

				rx_data = spi_mmio[CS0_RXDATA_FIFO];
				rxfifo_cnt--;

				if (rx_bitlen <= 8) {
					rx[0] = (u8) (rx_data & 0xff);
					rx_bitlen = 0;
				} else if (rx_bitlen <= 16) {
					rx[0] = (u8) ((rx_data >> 8) & 0xff);
					rx[1] = (u8) ((rx_data >> 0) & 0xff);
					rx_bitlen = 0;
				} else if (rx_bitlen <= 24) {
					rx[0] = (u8) ((rx_data >> 16) & 0xff);
					rx[1] = (u8) ((rx_data >> 8) & 0xff);
					rx[2] = (u8) ((rx_data >> 0) & 0xff);
					rx_bitlen = 0;
				} else {
					rx[0] = (u8) ((rx_data >> 24) & 0xff);
					rx[1] = (u8) ((rx_data >> 16) & 0xff);
					rx[2] = (u8) ((rx_data >> 8) & 0xff);
					rx[3] = (u8) ((rx_data >> 0) & 0xff);
					rx_bitlen = rx_bitlen - 32;
				}

				rx += 4;;
			}
		}
		if (tx_bitlen) {
			spi_txfifo_enq(spi_mmio, &tx_bitlen, &tx);
		}
	}

	/* Wait for done 
	 */
	while(!(spi_mmio[CS0_STATUS] & SPI_STATUS_DONE)) {};
	return 0;
}

int spi_xfer(struct spi_slave* slave,
		u32 bitlen, const void *dout,
		void *din, unsigned long flags) {

	int ret;

	while (bitlen != 0) {

		if (bitlen > 0x10000) {
			ret = spi_block_xfer(slave, 0x10000,
					dout, din, flags & ~(SPI_XFER_END));

			bitlen = bitlen - 0x10000;

			if (dout != NULL)
				dout = dout + 0x2000;
			if (din != NULL)
				din  = din  + 0x2000;
			if (ret)
				return ret;

		} else {
			ret = spi_block_xfer(slave, bitlen, dout, 
					din, flags);
			return ret;
		}
	}
	return 0;
}
