/*
 * Copyright 2003-2010 Netlogic Microsystems (“Netlogic”). All rights
 * reserved. Redistribution and use in source and binary forms, with 
 * or without modification, are permitted provided that the following 
 * conditions are met:
 *
 *	1. 	Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 *
 *	2. 	Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in
 *		the documentation and/or other materials provided with the
 *		distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * --------------------------------#NETL_2#--------------------------------
 */
#include <common.h>
#include <asm/arch/cpu.h>
#include "clock.h"

/* Time units in pico-seconds 
 */
#define PLL_REF_CLK_PS 7500

struct pll_settings
{
	u32 freq;
	u32 divf;
	u32 divr;
	u32 ext_div;
};

/* Freq, DivF, DivR, DFS
 */
static const struct pll_settings freq_lookup[] = 
{ 
	{400,  11,  0, 8},  /* 400Mhz */
	{433,  12,  0, 8},  /* 433Mhz */
	{466,  13,  0, 8},  /* 400Mhz */
	{500,  14,  0, 8},  /* 500MHz */
	{533,  15,  0, 8},  /* 533MHz */
	{566,  16,  0, 8},  /* 567MHz */
	{600,  17,  0, 8},  /* 600MHz */
	{633,  18,  0, 8},  /* 634MHz */
	{666,  14,  0, 6},  /* 667MHz */
	{700,  20,  0, 8},  /* 700MHz */
	{733,  21,  0, 8},  /* 734MHz */
	{766,  22,  0, 8},  /* 767MHz */
	{800,  11,  0, 4},  /* 800MHz */
	{866,  12,  0, 4},  /* 867MHz */
	{933,  13,  0, 4},  /* 934MHz */
	{1000, 14,  0, 4},  /* 1000MHz */
	{1066, 15,  0, 4},  /* 1066MHz */
	{1133, 16,  0, 4},  /* 1134MHz */
	{1200, 17,  0, 4},  /* 1200MHz */
	{1266, 18,  0, 4},  /* 1200MHz */
	{1333, 19,  0, 4},  /* 1333MHz */
	{1400, 20,  0, 4},  /* 1400MHz */
	{1466, 21,  0, 4},  /* 1400MHz */
	{1533, 22,  1, 4},  /* 1533MHz */
	{1600, 11,  0, 2},  /* 1600MHz */
	{1733, 12,  0, 2},  /* 1733MHz */
	{1866, 13,  0, 2},  /* 1867MHz */
	{2000, 14,  0, 2},  /* 2000MHz */
	{0   ,  0,  0, 0}
};

/* Return period in femto-seconds.(1e-15 s)
 */
static u32 get_pll_period(int divf,int divr) {

	u32 vco_fs; 		/* vco in femto-seconds */
	u32 pll_period_fs; 	/* pll output in femto-seconds */

	if (divr == 2) {
		printf ("Error! Illegal divr value %d!\n", divr);
		return 0;
	}
	if ((divf < 11) || (divf > 91)) {
		printf ("Error! Illegal divf value %d!\n", divf);
		return 0;
	}	

	vco_fs = (((PLL_REF_CLK_PS * 1000) * (divr+1))/(4 * (divf+1)));
	pll_period_fs  = vco_fs * 2; /* pll output is divided by 2 */

	return pll_period_fs;
}

/* Return frequency in KHz
 */
static u32 get_pll_freq(int divf,int divr)
{
	uint64_t khz_freq;
	uint64_t khz_mult   = (1000ULL * 1000ULL * 1000ULL * 1000ULL); /* 1e12 */
	u32 pll_period_fs 	= get_pll_period(divf,divr);

	khz_freq   = khz_mult/((uint64_t) pll_period_fs);
	
	return (u32) khz_freq;
}

/* Return frequency in KHz
 */
u32 get_cpu_freq(int cpu_num)
{
	volatile u32* mmio;
	u32 pwron_rst_reg;
	u32 core_dfs, divf, divr, dfs, core;
	
	mmio = (volatile u32 *) cpu_io_mmio(cpu_num/32,SYS);

	pwron_rst_reg = mmio[SYS_POWERONRESETCFG];
	core_dfs      = mmio[SYS_COREDFSDIVCTRL];
	core          = cpu_num >> 2;

	divf = SYS_PWRON_DIVF(pwron_rst_reg);
	divr = SYS_PWRON_DIVR(pwron_rst_reg);
	dfs  = SYS_CORE_DFS(core_dfs,core);

	return (get_pll_freq(divf,divr)/(dfs + 1));	
}

u32 get_dram_freq(int chip)
{
	volatile u32* mmio;
	u32 divf, divr, dfs;
	
	mmio = (volatile u32 *) cpu_io_mmio(chip, SYS);

	dfs  = SYS_DMC_DIV(mmio[SYS_SYSDFSDIVCTRL0]);
	divf = RD_DMC_PLL_DIVF(mmio[SYS_PLLCTRL]);
	divr = RD_DMC_PLL_DIVR(mmio[SYS_PLLCTRL]);
	
	return (get_pll_freq(divf,divr)/(dfs + 1));	
}

int find_freq_setting(u32 freq)
{
	int i = 0;

	while (freq_lookup[i].freq != 0) {
		if ((freq == freq_lookup[i].freq) |
		     (freq == (freq_lookup[i].freq + 1))) {
			return i;
		}
		i++;
	}
	return -1;
}

void disable_dram_clk(int chip)
{
	volatile u32* mmio = (volatile u32 *) cpu_io_mmio(chip, SYS);
	mmio[SYS_SYSDFSDISCTRL] = mmio[SYS_SYSDFSDISCTRL] | (SYS_DMC_DISABLE_MASK);	
}

void enable_dram_clk(int chip)
{
	volatile u32* mmio = (volatile u32 *) cpu_io_mmio(chip, SYS);
	mmio[SYS_SYSDFSDISCTRL] = mmio[SYS_SYSDFSDISCTRL] & ~(SYS_DMC_DISABLE_MASK);	
}

int8_t set_dram_freq(int chip, u32 freq)
{
	volatile u32* mmio;
	int index,dmc_div;

	mmio = (volatile u32 *) cpu_io_mmio(chip, SYS);

	if ((index = find_freq_setting(freq/1000)) == -1) {
		printf ("Error! Unable to set dram frequency %d MHz\n",
				freq/1000);
		return -1;
	}

	/* Disable DRAM clk */
	disable_dram_clk(chip);

	/* Setup correct external DFS divider */
	while (1) {

		dmc_div = (SYS_DMC_DIV(mmio[SYS_SYSDFSDIVCTRL0]) + 1);

		if (dmc_div == freq_lookup[index].ext_div) 
			break;

		if (dmc_div > freq_lookup[index].ext_div) {
			mmio[SYS_SYSDFSDIVDECCTRL] = SYS_DMC_DIV_DEC_MASK;
		} else {
			mmio[SYS_SYSDFSDIVINCCTRL] = SYS_DMC_DIV_INC_MASK;
		}
	}	

	/* Reset PLL and setup dividers */
	mmio[SYS_PLLCTRL] = ((mmio[SYS_PLLCTRL] & SYS_DMC_PLL_RMW_MASK) | 
			(SYS_DMC_PLL_RESET(1) | 
			 SYS_DMC_PLL_DIVR(freq_lookup[index].divr) | 
			 SYS_DMC_PLL_DIVF(freq_lookup[index].divf)));
	/* 1us for PLL reset */
	udelay(1);

	/* Remove PLL reset */
	mmio[SYS_PLLCTRL] = ((mmio[SYS_PLLCTRL] & SYS_DMC_PLL_RMW_MASK) | 
			(SYS_DMC_PLL_DIVR(freq_lookup[index].divr) | 
			 SYS_DMC_PLL_DIVF(freq_lookup[index].divf)));

	/* Wait 50us for PLL lock */
	udelay(50);

	/* Re-enable dram clock */
	enable_dram_clk(chip);

	return 0;
}
